// State of a Lindenmayer System

// Written by Bernie Roehl, November 1996

package lsystem;

import java.util.*;
import java.io.*;

import utility.*;

public class State {
    protected float x, y, z;
    protected Matrix rotmat;
    protected float angle;
    protected float length;
    protected int color;

    static float default_angle = 90;     // degrees
    static float default_length = 100;
    static int default_color = 1;

    int[] polydata = new int[1000];
    int polyindex = 0;
    boolean polyactive = false;

    public void setDefaultAngle(float a) { default_angle = a; }
    public void setDefaultLength(float l) { default_length = l; }
    public void setDefaultColor(int c) { default_color = c; }

    public State() {
        x = y = z = 0;
        angle = default_angle;
        length = default_length;
        color = default_color;
        rotmat = new Matrix();
    }

    public void cmd(char c, Emitter emitter) {
        switch (c) {
            case '+': rotmat = rotmat.multiply(Matrix.yrot(-angle)); break;
            case '-': rotmat = rotmat.multiply(Matrix.yrot(angle)); break;
            case '&': rotmat = rotmat.multiply(Matrix.xrot(angle)); break;
            case '^': rotmat = rotmat.multiply(Matrix.xrot(-angle)); break;
            case '\\':
            case '<': rotmat = rotmat.multiply(Matrix.zrot(angle)); break;
            case '/':
            case '>': rotmat = rotmat.multiply(Matrix.zrot(-angle)); break;
            case '|': rotmat = rotmat.multiply(Matrix.yrot(180)); break;
            case '%': rotmat = rotmat.multiply(Matrix.zrot(180)); break;
            case '"': ++length; break;
            case '\'': --length; break;
            case ';': ++angle; break;
            case ':': --angle; break;
            case 'c': ++color; break;
            case '*':
                      color = default_color;
                      length = default_length;
                      angle = default_angle;
                      break;
            case '{':
                      polyactive = true;
                      polyindex = 0;
                      polydata[polyindex++] = emitter.addVertex(x, y, z);
                      break;
            case '}':
                      emitter.addPolygon(polydata, polyindex, color);
                      polyactive = false;
                      break;
            case 'F':
            case 'Z':
                      if (polyactive == false) {
                          int n = emitter.addVertex(x, y, z);
                          advance((c == 'Z') ? 2 : 1);
                          emitter.addLine(n, emitter.addVertex(x, y, z), color);
                      }
                      else {
                          advance((c == 'Z') ? 2 : 1);
                          polydata[polyindex++] = emitter.addVertex(x, y, z);
                      }
                      break;
            case 'f':
            case 'z':
                      advance((c == 'z') ? 2 : 1);
                      if (polyactive)
                          polydata[polyindex++] = emitter.addVertex(x, y, z);
                      break;
            case 'g': advance(1); break;
            case '.':
                      if (polyactive) {
                         polydata[polyindex+1] = polydata[polyindex];
                         ++polyindex;
                      }
                      break;
            default: break;
        }
    } 

    private void advance(int divisor) {
        x += length * rotmat.getElement(0, 2) / divisor;
        y += length * rotmat.getElement(1, 2) / divisor;
        z += length * rotmat.getElement(2, 2) / divisor;
    }
}

