// A BSP Tree for spatial partitioning in VRML

// Written by Bernie Roehl, January 1997

package spatial;

import java.util.*;
import vrml.*;
import vrml.field.*;
import vrml.node.*;

public class BSPTree extends Script {
    BSPNode[] bspNodes;    // the non-leaf nodes of the BSP tree
    BSPLeaf[] bspLeaves;   // the leaf nodes of the BSP tree
    int oldleaf = -1;      // the leaf we were in the last time through
    int nchildren;         // number of child nodes (should be == number of leaves)
    SFInt32[] switches;    // the switch fields that control visibility
    boolean debug = false;

    public void initialize() {
        // find the whichChoice fields for all the child nodes
        MFNode children = (MFNode) getField("children");
        nchildren = children.getSize();
        Node[] nodes = new Node[nchildren];
        children.getValue(nodes);
        switches = new SFInt32[nchildren];
        for (int i = 0; i < nchildren; ++i)
            switches[i] = (SFInt32) nodes[i].getExposedField("whichChoice");
        // now load the non-leaf BSP tree nodes
        int n;
        MFString nodestring = (MFString) getField("bspNodes");
        n = nodestring.getSize();
        bspNodes = new BSPNode[n];
        String[] node_strings = new String[n];
        nodestring.getValue(node_strings);
        for (int i = 0; i < n; ++i)
            bspNodes[i] = new BSPNode(node_strings[i]);
        // now load the leaf nodes
        MFString leafstring = (MFString) getField("bspLeaves");
        n = leafstring.getSize();
        bspLeaves = new BSPLeaf[n];
        String[] leaf_strings= new String[n];
        leafstring.getValue(leaf_strings);
        for (int i = 0; i < n; ++i)
            bspLeaves[i] = new BSPLeaf(leaf_strings[i]);
    }

    public void processEvent(Event e) {
        if (e.getName().equals("location")) {  // user has moved
            // get their location and find out what leaf they're in
            ConstSFVec3f loc = (ConstSFVec3f) e.getValue();
            int leaf = whichLeaf(loc.getX(), loc.getY(), loc.getZ());
            if (leaf != oldleaf) {  // entered a new leaf
                if (debug) System.out.println("I'm in leaf " + leaf);
                // hide the old leaves
                for (int i = 0; i < nchildren; ++i)
                    switches[i].setValue(-1);
                // make the leaf node that we're in visible
                switches[leaf].setValue(0);
                // walk the list of leaves visible from this one
                Enumeration en = bspLeaves[leaf].getVislist().elements();
                if (debug) System.out.print("I can see ");
                while (en.hasMoreElements()) {
                    // find the integer leaf value
                    int el = ((Integer) en.nextElement()).intValue();
                    if (debug) System.out.print(" " + el);
                    // mark it as visible (whichChoice is zero)
                    switches[el].setValue(0);
                }
                if (debug) System.out.println();
                // keep track of the leaf we were in
                oldleaf = leaf;
             }
        }
    }

    // method for finding out which leaf node we're in
    int whichLeaf(float x, float y, float z) {
        int node = 0;
        // descend the tree (non-recursive!)
        while (node >= 0) {
            BSPNode thisnode = bspNodes[node];
            if (thisnode.inFront(x, y, z))
                node = thisnode.getLeft();
            else
                node = thisnode.getRight();
        }
        return -node - 1;  // leaf nodes count -1, -2, -3 ...
    }

}

